var eztv = require('eztv-api');
var chalk = require('chalk');
var inquirer = require('inquirer');
var Datastore = require('nedb');
var Rx = require('rx');
var numeral = require('numeral');
var clivas = require('clivas');
var torrentStream = require('torrent-stream');

var dbShows = new Datastore({ filename: 'data/shows.db' });
dbShows.loadDatabase(checkError);

var checkError = function(error) {
  if (error) {
    console.log(chalk.red(error));
    process.exit(1);
  }
};

var mainMenu = function() {
  inquirer.prompt([{
    type: 'list',
    name: 'menu',
    message: 'What do you want to do?',
    choices: [
      'Download an episode',
      'Add a serie',
      'Delete a serie',
      'Quit'
    ]
  }], function(answer) {
      if (answer.menu === 'Add a serie') {
        addSerie();
      } else if (answer.menu === 'Download an episode') {
        downloadEpisode();
      } else if (answer.menu === 'Delete a serie') {
        deleteSerie();
      } else if (answer.menu === 'Quit') {
        process.exit(0);
      }
    }
  );
};

var downloadEpisode = function() {
  Rx.Observable.create(function(observer) {
    dbShows.find({}, function (err, docs) {
      if (err) {
        observer.onError(err);
      } else {
        observer.onNext(docs);
      }
      observer.onCompleted();
    });
  })
  .map(function(showList) {
    var choices = [];
    for (var i = 0; i < showList.length; i++) {
      var ls = '' + showList[i].lastSeenSeasonNumber;
      var le = '' + showList[i].lastSeenEpisodeNumber;
      if (ls.length == 1) ls = '0' + ls;
      if (le.length == 1) le = '0' + le;
      choices.push({
        name: showList[i].title + ' (last seen S'+ls+'E'+le+')',
        value: showList[i]
      });
    }
    return choices;
  })
  .flatMap(flatMapChoose)
  .map(mapAnswer)
  .flatMap(getEpisodesBySerie)
  .flatMap(function (result) {
    if (result.episodes) {
      return Rx.Observable
        .from(result.episodes)
        .map(function (episode) {
          episode.serie = result.serie;
          return episode;
        })
        .filter(function (episode) {
          return (episode.seasonNumber > result.serie.lastSeenSeasonNumber) || (episode.seasonNumber == result.serie.lastSeenSeasonNumber && episode.episodeNumber > result.serie.lastSeenEpisodeNumber);
        });
    } else {
      return Rx.Observable.just(null);
    }
  })
  .reduce(reduceItemsToList, [])
  .map(function(episodeList) {
    var choices = [];
    for (var i = 0; i < episodeList.length; i++) {
      choices.push({
        name: episodeList[i].title,
        value: episodeList[i]
      });
    }
    return choices;
  })
  .flatMap(flatMapChoose)
  .map(mapAnswer)
  .flatMap(downloadEpisodeMagnet)
  .flatMap(updateSerieWithLastEpisode)
  .subscribe(
    function (x) {
      if (!x) {
        console.log(chalk.green('No new episodes found'));
        mainMenu();
      } else {
        console.log(chalk.green('Episode has been downloaded'));
        mainMenu();
      }
    }
  );
};

var deleteSerie = function() {
  Rx.Observable.create(function(observer) {
    dbShows.find({}, function (err, docs) {
      if (err) {
        observer.onError(err);
      } else {
        observer.onNext(docs);
      }
      observer.onCompleted();
    });
  })
  .map(function(showList) {
    var choices = [];
    for (var i = 0; i < showList.length; i++) {
      var ls = '' + showList[i].lastSeenSeasonNumber;
      var le = '' + showList[i].lastSeenEpisodeNumber;
      if (ls.length == 1) ls = '0' + ls;
      if (le.length == 1) le = '0' + le;
      choices.push({
        name: showList[i].title + ' (last seen S'+ls+'E'+le+')',
        value: showList[i]
      });
    }
    return choices;
  })
  .flatMap(flatMapChoose)
  .map(mapAnswer)
  .flatMap(function(serie) {
    return Rx.Observable.create(function(observer) {
      dbShows.remove({_id: serie._id}, {}, function (err, result) {
        if (err) {
          observer.onError(err);
        } else {
          observer.onNext(result);
        }
        observer.onCompleted();
      });
    });
  })
  .subscribe(
    function (x) {
      console.log(chalk.green('This has been deleted from your shows'));
      mainMenu();
    }
  );
};

var addSerie = function() {
  inquirer.prompt([{
    type: 'input',
    name: 'name',
    message: 'What\'s is the name of this serie ?'
  }]).process
  .map(mapAnswer)
  .flatMap(getSerieByName)
  .map(function(showList) {
    var choices = [];
    for (var i = 0; i < showList.length; i++) {
      choices.push({
        name: showList[i].title,
        value: showList[i]
      });
    }
    return choices;
  })
  .flatMap(flatMapChoose)
  .map(mapAnswer)
  .flatMap(function (show) {
    if (show) {
      return Rx.Observable.create(function(observer) {
        setTimeout(function () {
          inquirer.prompt([{
            type: 'input',
            name: 'seasonNumber',
            message: 'Type the season number which you have seen the last episode :'
            },{
              type: 'input',
              name: 'episodeNumber',
              message: 'Type the last seen episode number :'
            }], function (answers) {
            show.lastSeenSeasonNumber = parseInt(answers.seasonNumber);
            show.lastSeenEpisodeNumber = parseInt(answers.episodeNumber);
            observer.onNext(show);
            observer.onCompleted();
          });
        });
      });
    } else {
      return Rx.Observable.just(null);
    }
  })
  .subscribe(
    function (show) {
      if (show) {
        dbShows.insert(show, function (err, newDoc) {
          console.log(chalk.green(show.title + ' saved to your shows'));
        });
      } else {
        console.log(chalk.green('This show was not found'));
      }
      mainMenu();
    }
  );
};

var downloadEpisodeMagnet = function(episode) {
  if (episode) {
    var t = episode.show.replace(/\ /g, '.');
    var s = '' + episode.seasonNumber;
    var e = '' + episode.episodeNumber;
    if (s.length == 1) s = '0' + s;
    if (e.length == 1) e = '0' + e;
    var options = {
      path: 'download/' + t + '.S' + s + 'E' + e + '.mkv'
    };

    return Rx.Observable.create(function(observer) {
      var engine = torrentStream(episode.magnet, options);
      var fileDownloaded = null;
      var verified = 0;
      var invalid = 0;
      var hotswaps = 0;

      engine.on('idle', function() {
        engine.remove(true, function () {
          observer.onNext(episode);
          observer.onCompleted();
        });
      });

      engine.on('verify', function () {
        verified++;
      });

      engine.on('invalid-piece', function () {
        invalid++;
      });

      engine.on('hotswap', function () {
        hotswaps++;
      });

      engine.on('download', function(a) {
        var swarm = engine.swarm;
        var unchoked = swarm.wires.filter(function (wire) {
          return !wire.peerChoking;
        });
        var percent = Math.round(10000 * swarm.downloaded / fileDownloaded.length) / 100;
        clivas.clear(); // clears the canvas
        clivas.line('{green:File} {bold:' + fileDownloaded.name + ' (' + bytes(fileDownloaded.length) + ')}');
        clivas.line('{green:Downloaded} {bold:' + bytes(swarm.downloaded) + ' [' + percent + '%]} {green:- Uploaded} {bold:' + bytes(swarm.uploaded) + '}');
        clivas.line('{green:Speed} {bold:' + bytes(swarm.downloadSpeed()) + '/s} {green:from} {bold:' + unchoked.length + '/' + swarm.wires.length + '} {green:peers}');
        clivas.line('{green:Verified} {bold:' + verified + '} {green:pieces and received} {bold:' + invalid + '} {green:invalid pieces}');
        clivas.line('{green:Peer queue size is} {bold:' + swarm.queued + '}');
        clivas.flush();
      });

      engine.on('ready', function() {
        var maxLength = 0;
        engine.files.forEach(function(file) {
          if (file.length > maxLength) {
            maxLength = file.length;
            fileDownloaded = file;
          }
          fileDownloaded.select();
          fileDownloaded.createReadStream();
        });
      });
    });
  } else {
    return Rx.Observable.just(null);
  }
};

var updateSerieWithLastEpisode = function(episode) {
  if (episode) {
    return Rx.Observable.create(function(observer) {
      dbShows.update({id: episode.serie.id}, { $set: {
        lastSeenSeasonNumber: episode.seasonNumber,
        lastSeenEpisodeNumber: episode.episodeNumber
      }}, function (error, newDoc) {
        if (error) {
          observer.onError(error);
        } else {
          observer.onNext(newDoc);
        }
        observer.onCompleted();
      });
    });
  } else {
    return Rx.Observable.just(null);
  }
};

var getSerieByName = function(name) {
  return Rx.Observable.create(function(observer) {
    eztv.getShows({query: name}, function(error, results) {
      if (error) {
        observer.onError(error);
      } else {
        observer.onNext(results);
      }
      observer.onCompleted();
    });
  });
};

var getEpisodesBySerie = function(serie) {
  return Rx.Observable.create(function(observer) {
    eztv.getShowEpisodes(serie.id, function(error, result) {
      if (error) {
        observer.onError(error);
      } else {
        result.serie = serie;
        observer.onNext(result);
      }
      observer.onCompleted();
    });
  });
};

var flatMapChoose = function(choices) {
  if (choices.length > 1) {
    return inquirer.prompt([{
      type: 'list',
      name: 'question',
      message: 'Choose a result',
      paginated: false,
      choices: choices
    }]).process;
  } else if (choices.length > 0) {
    return Rx.Observable.just({name : 'question', answer: choices[0].value});
  } else {
    return Rx.Observable.just({name : 'question', answer: null});
  }
};

var mapAnswer = function(answers) {
  return answers.answer;
};

var reduceItemsToList = function (list, item) {
  if (item) {
    list.push(item);
  }
  return list;
};

var bytes = function (num) {
  return numeral(num).format('0.0b');
};

mainMenu();
